//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#include "stdafx.h"
#include <io.h>
#include "emule.h"
#include "KnownFileList.h"
#include "KnownFile.h"
#include "opcodes.h"
#include "Preferences.h"
#include "SafeFile.h"
#include "OtherFunctions.h"
#ifndef _CONSOLE
#include "emuledlg.h"
#endif
#include "DownloadQueue.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


#define KNOWN_MET_FILENAME	_T("known.met")


CKnownFileList::CKnownFileList()
{
	accepted = 0;
	requested = 0;
	transferred = 0;
	m_nLastSaved = ::GetTickCount();
	Init();
}

CKnownFileList::~CKnownFileList()
{
	Clear();
}

bool CKnownFileList::Init()
{
	CString fullpath=theApp.glob_prefs->GetConfigDir();
	fullpath.Append(KNOWN_MET_FILENAME);
	CSafeBufferedFile file;
	CFileException fexp;
	if (!file.Open(fullpath,CFile::modeRead|CFile::osSequentialScan|CFile::typeBinary, &fexp)){
		if (fexp.m_cause != CFileException::fileNotFound){
			CString strError(_T("Failed to load ") KNOWN_MET_FILENAME _T(" file"));
			TCHAR szError[MAX_CFEXP_ERRORMSG];
			if (fexp.GetErrorMessage(szError, ARRSIZE(szError))){
				strError += _T(" - ");
				strError += szError;
			}
			AddLogLine(true, _T("%s"), strError);
		}
		return false;
	}
	setvbuf(file.m_pStream, NULL, _IOFBF, 16384);

	CSingleLock sLock(&list_mut);
	try {
		uint8 header;
		file.Read(&header,1);
		if (header != MET_HEADER){
			file.Close();
			return false;
		}
		
		uint32 RecordsNumber;
		file.Read(&RecordsNumber,4);
		sLock.Lock();
		for (uint32 i = 0; i < RecordsNumber; i++) {
			CKnownFile* pRecord =  new CKnownFile();
			if (!pRecord->LoadFromFile(&file)){
				TRACE("*** Failed to load entry %u (name=%s  hash=%s  size=%u  parthashs=%u expected parthashs=%u) from known.met\n", i, 
					pRecord->GetFileName(), md4str(pRecord->GetFileHash()), pRecord->GetFileSize(), pRecord->GetHashCount(), pRecord->GetED2KPartHashCount());
				delete pRecord;
				continue;
			}
			Add(pRecord);
		}
		sLock.Unlock();
		file.Close();
	}
	catch(CFileException* error){
		if (error->m_cause == CFileException::endOfFile)
			AddLogLine(true,GetResString(IDS_ERR_SERVERMET_BAD));
		else{
			char buffer[MAX_CFEXP_ERRORMSG];
			error->GetErrorMessage(buffer,MAX_CFEXP_ERRORMSG);
			AddLogLine(true,GetResString(IDS_ERR_SERVERMET_UNKNOWN),buffer);
		}
		error->Delete();
		return false;
	}

	return true;
}

void CKnownFileList::Save()
{
	if (theApp.glob_prefs->GetLogFileSaving())
		DEBUG_ONLY(AddDebugLogLine(false, "Saved KnownFileList"));
	m_nLastSaved = ::GetTickCount(); 
	CString fullpath=theApp.glob_prefs->GetConfigDir();
	fullpath += KNOWN_MET_FILENAME;
	CSafeBufferedFile file;
	CFileException fexp;
	if (!file.Open(fullpath, CFile::modeWrite|CFile::modeCreate|CFile::typeBinary, &fexp)){
		CString strError(_T("Failed to save ") KNOWN_MET_FILENAME _T(" file"));
		TCHAR szError[MAX_CFEXP_ERRORMSG];
		if (fexp.GetErrorMessage(szError, ARRSIZE(szError))){
			strError += _T(" - ");
			strError += szError;
		}
		AddLogLine(true, _T("%s"), strError);
		return;
	}
	setvbuf(file.m_pStream, NULL, _IOFBF, 16384);

	CSingleLock sLock(&list_mut);
	try{
		uint8 ucHeader = MET_HEADER;
		file.Write(&ucHeader, 1);

		uint32 RecordsNumber = GetCount();
		file.Write(&RecordsNumber, 4);
		for (uint32 i = 0; i < RecordsNumber; i++)
			ElementAt(i)->WriteToFile(&file);
		if (theApp.glob_prefs->GetCommitFiles() >= 2 || (theApp.glob_prefs->GetCommitFiles() >= 1 && !theApp.emuledlg->IsRunning())){
			file.Flush(); // flush file stream buffers to disk buffers
			if (_commit(_fileno(file.m_pStream)) != 0) // commit disk buffers to disk
				AfxThrowFileException(CFileException::hardIO, GetLastError(), file.GetFileName());
		}
		file.Close();
	}
	catch(CFileException* error){
		CString strError(_T("Failed to save ") KNOWN_MET_FILENAME _T(" file"));
		TCHAR szError[MAX_CFEXP_ERRORMSG];
		if (error->GetErrorMessage(szError, ARRSIZE(szError))){
			strError += _T(" - ");
			strError += szError;
		}
		AddLogLine(true, _T("%s"), strError);
		error->Delete();
	}
}

void CKnownFileList::Clear()
{
	for (int i = 0; i < GetSize();i++)
		safe_delete(this->ElementAt(i));
	RemoveAll();
	SetSize(0);
}

CKnownFile* CKnownFileList::FindKnownFile(LPCTSTR filename,uint32 in_date,uint32 in_size)
{
	for (int i = 0;i < GetCount();i++){
		CKnownFile* cur_file = ElementAt(i);
		if (cur_file->GetFileDate() == in_date && cur_file->GetFileSize() == in_size && (!_tcscmp(filename,cur_file->GetFileName())))
			return cur_file;
	}
	return 0;
}

void CKnownFileList::SafeAddKFile(CKnownFile* toadd)
{
	CSingleLock sLock(&list_mut,true);
	Add(toadd);
	sLock.Unlock();
}

CKnownFile* CKnownFileList::FindKnownFileByID(const uchar* hash){
	for (int i = 0; i < GetCount(); i++){
		CKnownFile* pCurKnownFile = ElementAt(i);
		if (!md4cmp(pCurKnownFile->GetFileHash(), hash))
			return pCurKnownFile;
	}
	return NULL;
}

void CKnownFileList::Process()
{
	if ( ::GetTickCount() - m_nLastSaved > MIN2MS(11))
		this->Save();
}

bool CKnownFileList::IsKnownFile(void* pToTest){
	for (int i = 0; i < GetCount(); i++){
		if (ElementAt(i) == pToTest)
			return true;
	}
	return false;
}

// emulEspaa: Added by [TPT]-MoNKi [MoNKi: -Check already downloaded files-]
// returns:
//		1 if a file was found
//		2 if more than 1 file found or only the name is equal.
//		0 if no file found
int CKnownFileList::CheckAlreadyDownloadedFile(const uchar* hash, CString filename, CArray<CKnownFile*,CKnownFile*> *files)
{
	CKnownFile* sameFile=NULL;
	CArray<CKnownFile*,CKnownFile*> filesFound;

	if(hash==NULL && !filename.IsEmpty()){
		for (int i = 0;i < GetCount() && !sameFile;i++){
			CKnownFile* cur_file = ElementAt(i);
			if (filename == cur_file->GetFileName() && !cur_file->IsPartFile())
				filesFound.Add(cur_file);
		}
		if(files){
			if(!files->IsEmpty()) files->RemoveAll();
			files->Append(filesFound);
		}
		if(!filesFound.IsEmpty())
			return 2;
		else
			return 0;
	}
	else if(hash!=NULL){
		CKnownFile* cur_file = FindKnownFileByID(hash);
		if(cur_file && !cur_file->IsPartFile()){
			filesFound.Add(cur_file);
			if(files){
				if(!files->IsEmpty()) files->RemoveAll();
				files->Add(cur_file);
			}
			return 1;
		}
		else
			return 0;
	}
	return 0;
}

//Returns:
//	true if you can download it
bool CKnownFileList::CheckAlreadyDownloadedFileQuestion(const uchar* hash, CString filename){
	CArray<CKnownFile *,CKnownFile*> filesFound;
	int ret;

	if(theApp.downloadqueue->IsFileExisting(hash)){
		return false;
	}

	ret=CheckAlreadyDownloadedFile(hash, "", &filesFound);
	if(ret==0)
		ret=CheckAlreadyDownloadedFile(NULL,filename, &filesFound);

	if(ret!=0)
	{
		CKnownFile* curFile=NULL;
		CString msg;
		if(ret==1){
			msg = GetResString(IDS_DOWNHISTORY_CHECK1);
		}
		else {
			msg.Format(GetResString(IDS_DOWNHISTORY_CHECK2), filesFound.GetCount(), filename);
		}

		for(int i=0;i<filesFound.GetCount();i++){
			CKnownFile *cur_file = filesFound.GetAt(i);
			CString sData;

			msg+=cur_file->GetFileName() + "\n";
			sData.Format(GetResString(IDS_DL_SIZE) + ": %u, " + GetResString(IDS_FILEID) + ": %s", cur_file->GetFileSize(), EncodeBase16(cur_file->GetFileHash(),16));
			msg+=sData;
			if(!cur_file->GetFileComment().IsEmpty()) //Add comment
				msg+="\n" + GetResString(IDS_COMMENT) + ": \"" + cur_file->GetFileComment() + "\"";
			msg+="\n\n";
		}
		msg += GetResString(IDS_DOWNHISTORY_CHECK3);
		if(MessageBox(NULL, msg, GetResString(IDS_DOWNHISTORY),MB_YESNO|MB_ICONQUESTION)==IDYES)
			return true;
		else
			return false;
	}
	else
	{
		return true;
	}
}
//End emulEspaa

//emulEspaa: Added by [TPT]-MoNKi [MoNKi: -Downloaded History-]
CArray<CKnownFile*,CKnownFile*> * CKnownFileList::GetDownloadedFiles(){
	CArray<CKnownFile*,CKnownFile*> * filesFound;

	filesFound = new CArray<CKnownFile*,CKnownFile*>;

	for (int i = 0;i < GetCount();i++){
		CKnownFile* cur_file = ElementAt(i);
		if (!cur_file->IsShared()) filesFound->Add(cur_file);
	}
	return filesFound;
}

bool CKnownFileList::RemoveKnownFile(CKnownFile *toRemove){
	CSingleLock sLock(&list_mut,true);

	for(int i=0; i<GetCount(); i++){
		CKnownFile *foundFile = GetAt(i);
		if (foundFile == toRemove){
			RemoveAt(i);
			delete toRemove;
			sLock.Unlock();
			return true;
		}
	}
	sLock.Unlock();
	return false;
}

void CKnownFileList::ClearHistory(){
	CSingleLock sLock(&list_mut,true);

	for(int i=GetCount()-1; i>=0; i--){
		CKnownFile *cur_file = GetAt(i);
		if (!cur_file->IsShared()){
			RemoveAt(i);
			delete cur_file;
		}
	}
	sLock.Unlock();
}
//End emulEspaa

//emulEspaa: Added By [TPT]-MoNKi [-xrmb\MoNKi: MergeKnown-]
void CKnownFileList::RemoveDuplicates(){
	CSingleLock sLock(&list_mut,true);

	int items=GetCount();

	//--- go thru all items ---
	for(int i=0; i<items; i++)
	{
		//--- compare with all other items ---
		for(int j=i+1; j<items; j++)
		{
			//--- same id? and same size... maybe we should compare the hashsets too? ---
			if(	ElementAt(i)->GetFileSize()==ElementAt(j)->GetFileSize() &&
				md4cmp(ElementAt(i)->GetFileHash(), ElementAt(j)->GetFileHash())==0
				)
			{
				//--- one file can be shared ---
				if(ElementAt(i)->IsShared() && ElementAt(j)->IsShared())
				{
					sLock.Unlock();
					return;
				}

				//--- if second file is shared, swap them ---
				if(ElementAt(j)->IsShared())
				{
					CKnownFile *t=ElementAt(i);
					SetAt(i, ElementAt(j));
					SetAt(j, t);
				}

				ElementAt(i)->statistic.AddTransferred(0, ElementAt(j)->statistic.GetAllTimeTransferred()); // Pichuei comprobar si es correcto empezar el start a 0

				//--- remove second item ---
				CKnownFile	*d=ElementAt(j);
				
				ASSERT(!d->IsPartFile());
				
				RemoveAt(j);
				delete d;
				items--;
				j--;
			}
		}

	}
	sLock.Unlock();
}
//End emulEspaa
